import urllib.request
import urllib.error
import argparse
import json
import sys
import os
from subprocess import check_call, CalledProcessError
from importlib.resources import files, as_file
from .parse.parser import WCAParser
from .sema.ast import WCARegulations, WCAGuidelines, WCAStates, Ruleset
from .codegen.cghtml import WCADocumentHtml
from .codegen.cghtmltopdf import WCADocumentHtmlToPdf
from .codegen.cgjson import WCADocumentJSON
from .codegen.merger import merge_ast
from .version import __version__

REGULATIONS_FILENAME = "wca-regulations.md"
GUIDELINES_FILENAME = "wca-guidelines.md"
STATES_FILENAME = "wca-states.md"
LANGUAGES_JSON_URL = "https://raw.githubusercontent.com/thewca/wca-regulations-compiler/main/wrc/data/languages.json"
URLLIB_REQUEST_TIMEOUT = 8


def parse_states(states):
    states_as_str = None
    astreg = None
    errors = []
    warnings = []
    if states:
        with open(states, encoding='utf-8') as states_file:
            states_as_str = states_file.read()
    parser = WCAParser()
    if states_as_str:
        astreg, errors, warnings = parser.parse(states_as_str, WCAStates)
    return (astreg, None, errors, warnings)


def parse_regulations_guidelines(reg, guide):
    reg_as_str = None
    guide_as_str = None
    astreg = None
    astguide = None
    errors = []
    warnings = []
    if reg:
        reg_as_str = get_file_as_str(reg)
    if guide:
        guide_as_str = get_file_as_str(guide)

    parser = WCAParser()
    if reg_as_str:
        astreg, errors_reg, warnings_reg = parser.parse(reg_as_str, WCARegulations)
        errors.extend(errors_reg)
        warnings.extend(warnings_reg)
    if guide_as_str:
        astguide, errors_guide, warnings_guide = parser.parse(guide_as_str, WCAGuidelines)
        errors.extend(errors_guide)
        warnings.extend(warnings_guide)
    return (astreg, astguide, errors, warnings)


def get_file_as_str(file) -> str:
    """
    * TAB -> 4 spaces.
    * Add EOL to EOF if missing.
    :return: file as string.
    """
    with open(file, encoding='utf-8') as md_file:
        file_as_str = md_file.read()
    # FIXME: do we want to just remove tabs from regs?
    file_as_str = file_as_str.replace("\t", "    ")

    # Add EOL to EOF if missing.
    if file_as_str[-1] != '\n':
        file_as_str += '\n'

    return file_as_str


def output(result_tuple, outputs, output_dir):
    output_filename = None
    for content, filename in zip(result_tuple, outputs):
        mode = 'w'
        if output_filename == output_dir + "/" + filename:
            mode = 'a'
        else:
            output_filename = output_dir + "/" + filename
        with open(output_filename, mode + '+', encoding='utf-8') as output_file:
            output_file.write(content)
            print("Successfully written the content to " + output_filename)


def generate(backend_class, inputs, outputs, options, parsing_method, post_process=None, merged=False):
    astreg, astguide, errors, warnings = parsing_method(*inputs)
    if len(errors) + len(warnings) == 0:
        print(("Compiled document, generating " +
               backend_class.name + "..."))
        languages_options = languages(False)[options.language]
        if merged:
            languages_options["pdf"] += ".merged"
        if backend_class is WCADocumentHtml:
            # Other subclasses of CGDocument do not have a "merged" constructor parameter.
            cg_instance = backend_class(options.git_hash, options.language,
                                        languages_options["pdf"], merged=merged)
        else:
            cg_instance = backend_class(options.git_hash, options.language,
                                        languages_options["pdf"])
        if merged:
            # Merge both ASTs and do not visit astguide.
            astreg = merge_ast(astreg, astguide, languages_options)
            result_tuple = cg_instance.emit(astreg, None)
        else:
            result_tuple = cg_instance.emit(astreg, astguide)

        output(result_tuple, outputs, options.output)
        if post_process:
            post_process(outputs, options.output, languages_options)
    return (errors, warnings)


def html_to_pdf(tmp_filenames, output_directory, lang_options):
    input_html = output_directory + "/" + tmp_filenames[0]
    wkthml_cmd = ["wkhtmltopdf"]
    # Basic margins etc
    wkthml_cmd.extend(["--margin-left", "18"])
    wkthml_cmd.extend(["--margin-right", "18"])
    wkthml_cmd.extend(["--page-size", "Letter"])
    # Header and Footer
    header_resource = files("wrc").joinpath("data/header.html")
    footer_resource = files("wrc").joinpath("data/footer.html")
    header_file = as_file(header_resource)
    footer_file = as_file(footer_resource)
    wkthml_cmd.extend(["--header-html", str(header_file)])
    wkthml_cmd.extend(["--footer-html", str(footer_file)])
    wkthml_cmd.extend(["--header-spacing", "8"])
    wkthml_cmd.extend(["--footer-spacing", "8"])
    wkthml_cmd.extend(["--enable-local-file-access"])
    wkthml_cmd.append(input_html)
    wkthml_cmd.append(output_directory + "/" + lang_options['pdf'] + '.pdf')
    try:
        check_call(wkthml_cmd)
        print("Successfully generated pdf file!")
        print("Cleaning temporary file (%s)..." % input_html)
        os.remove(input_html)
    except CalledProcessError as err:
        print("Error while generating pdf:")
        print(err)
        sys.exit(1)
    except OSError as err:
        print("Error when running command \"" + " ".join(wkthml_cmd) + "\"")
        print(err)
        sys.exit(1)


def output_diff(submitted, reference):
    rules_visitor = Ruleset()
    set_submitted = rules_visitor.get(submitted)
    set_reference = rules_visitor.get(reference)
    unexpected = set_submitted - set_reference
    missing = set_reference - set_submitted
    if len(unexpected) > 0:
        print(("/!\\ These numbers are in the translation file, "
               "but not in the reference one: {%s}" % ', '.join(sorted(unexpected))))
    if len(missing) > 0:
        print(("/!\\ These numbers are in the reference file, "
               "but not in the translation one: {%s}" % ', '.join(sorted(missing))))
    return len(unexpected) + len(missing)


def generate_diff(input_ast_reg, input_ast_guide, options):
    ref_regulations, ref_guidelines = files_from_dir(options.diff)
    parse_ref = parse_regulations_guidelines(ref_regulations, ref_guidelines)
    astreg_ref, astguide_ref, errors, warnings = parse_ref
    if len(errors) + len(warnings) != 0:
        print("Couldn't compile reference Regulations and Guidelines")
    else:
        print("All checks passed!")
        diffs = 0
        if input_ast_reg and astreg_ref:
            diffs += output_diff(input_ast_reg, astreg_ref)
        elif input_ast_reg:
            print("No reference to compare the intput Regulations to")

        if input_ast_guide and astguide_ref:
            diffs += output_diff(input_ast_guide, astguide_ref)
        elif input_ast_guide:
            print("No reference to compare the input Guidelines to")
        if diffs == 0:
            print("Input file(s) and reference file(s) matched!")
        else:
            errors.append("Translation and reference did not match!")
    return (errors, warnings)


def files_from_dir(file_or_directory):
    regulations = None
    guidelines = None
    if os.path.isdir(file_or_directory):
        regulations = file_or_directory + "/" + REGULATIONS_FILENAME
        guidelines = file_or_directory + "/" + GUIDELINES_FILENAME
        if (not os.path.isfile(regulations) or
                not os.path.isfile(guidelines)):
            print(("Error: the directory '%s' must contain both the Regulations "
                   "and Guidelines files." % file_or_directory))
            sys.exit(1)
    elif os.path.isfile(file_or_directory):
        if file_or_directory.endswith(REGULATIONS_FILENAME):
            regulations = file_or_directory
        elif file_or_directory.endswith(GUIDELINES_FILENAME):
            guidelines = file_or_directory
        else:
            print("Error: couldn't detect if the input file are Regulations or Guidelines.")
            sys.exit(1)
    else:
        print(f"Warning: {file_or_directory} is not a file or a directory. Ignoring.")
        sys.exit(0)
    return regulations, guidelines


def check_output(directory):
    if not os.path.isdir(directory):
        print("Error: output is not a directory.")
        sys.exit(1)


def languages(display=True):
    # First try to fetch the languages.json file from the compiler's repository.
    try:
        with urllib.request.urlopen(LANGUAGES_JSON_URL, timeout=URLLIB_REQUEST_TIMEOUT) as req:
            languages_json_str = req.read().decode(req.headers.get_content_charset())
    except (urllib.error.HTTPError, urllib.error.URLError) as e:
        print(f"Warning: could not fetch the latest languages.json file, a fallback will be used. Reason: {e.reason}")

        # Get information about languages from the config file (tex encoding, pdf filename, etc)
        languages_json_str = files(__package__).joinpath("data/languages.json").read_text(encoding="utf-8")

    languages_info = json.loads(languages_json_str)

    if display:
        print(" ".join([key for key in list(languages_info.keys()) if key != "english"]))
        sys.exit(0)

    return languages_info


def check_states_file(file_or_directory):
    if not os.path.isfile(file_or_directory) or not file_or_directory.endswith(STATES_FILENAME):
        print("Error: input file is not as expected..")
        sys.exit(1)


def build_common_option(argparser):
    argparser.add_argument('-o', '--output', default='build/', help='Output directory')
    argparser.add_argument('input', help='Input file or directory')
    argparser.add_argument('-g', '--git-hash', default='unknown',
                           help='Git hash corresponding to the files')
    argparser.add_argument('-l', '--language', default='english', help='Language of the file')


def handle_errors_and_warnings(errors, warnings):
    # If some errors or warnings have been detected, output them
    if len(errors) + len(warnings) != 0:
        print("Couldn't compile file, the following occured:")
        for err in errors:
            print(" - Error: " + err)
        for warn in warnings:
            print(" - Warning: " + warn)
        sys.exit(1)


def states():
    argparser = argparse.ArgumentParser()
    build_common_option(argparser)
    argparser.add_argument('--target', help='Select target output kind',
                           choices=['check', 'json'])
    options = argparser.parse_args()

    check_states_file(options.input)

    if options.target == "json":
        check_output(options.output)
        errors, warnings = generate(WCADocumentJSON,
                                    [options.input],
                                    ["states.json"],
                                    options,
                                    parse_states)
    elif options.target == "check":
        print("Checking input file(s)...")
        astreg, astguide, errors, warnings = parse_states(options.input)
        if len(errors) + len(warnings) == 0:
            print("All checks passed!")
    else:
        print("Nothing to do, exiting...")
        sys.exit(0)

    handle_errors_and_warnings(errors, warnings)


def run():
    argparser = argparse.ArgumentParser()
    action_group = argparser.add_mutually_exclusive_group()
    action_group.add_argument('--target', help='Select target output kind',
                              choices=['latex', 'pdf', 'html', 'check', 'json'])
    action_group.add_argument('--diff', help='Diff against the specified file')
    action_group.add_argument('-v', '--version', action='version',
                              version=__version__)
    argparser.add_argument('-m', '--merged',
                           help='Merge the Regulations and Guidelines into a single document', action='store_true')
    build_common_option(argparser)

    options = argparser.parse_args()

    if not options.diff and not options.target:
        print("Nothing to do, exiting...")
        sys.exit(0)

    input_regulations, input_guidelines = files_from_dir(options.input)

    errors = []
    warnings = []

    if options.target == "html":
        check_output(options.output)
        if options.merged:
            output_files = ["full.html.erb"]
        else:
            output_files = ["index.html.erb", "guidelines.html.erb"]
        errors, warnings = generate(WCADocumentHtml,
                                    (input_regulations, input_guidelines),
                                    output_files,
                                    options, parse_regulations_guidelines, merged=options.merged)
    elif options.target == "pdf":
        check_output(options.output)
        if not input_regulations:
            print("Error: At least the Regulations are needed "
                  "to generate the pdf file.")
            sys.exit(1)
        errors, warnings = generate(WCADocumentHtmlToPdf,
                                    (input_regulations, input_guidelines),
                                    ["regulations_tmp.html", "regulations_tmp.html"],
                                    options, parse_regulations_guidelines,
                                    html_to_pdf, merged=options.merged)
        # errors, warnings = generate_htmltopdf(input_regulations, input_guidelines, options)
    elif options.target == "json":
        check_output(options.output)
        errors, warnings = generate(WCADocumentJSON,
                                    (input_regulations, input_guidelines),
                                    ["wca-regulations.json"],
                                    options, parse_regulations_guidelines, merged=options.merged)
    elif options.target == "check" or options.diff:
        print("Checking input file(s)...")
        astreg, astguide, errors, warnings = parse_regulations_guidelines(input_regulations,
                                                                          input_guidelines)
        if len(errors) + len(warnings) == 0:
            print("All checks passed!")
            if options.diff:
                print("Checking reference file(s) for diff")
                errors, warnings = generate_diff(astreg, astguide, options)

    handle_errors_and_warnings(errors, warnings)


if __name__ == '__main__':
    run()



''' Backend for PDF using html. '''
import os.path
from importlib.resources import files, as_file
from wrc.sema.ast import Rule, LabelDecl, Article
from wrc.codegen.cghtml import WCADocumentHtml

CSS_FONTS = '''
@font-face {{
    font-family: "Computer Modern";
    src: url("{normal}");
    font-weight: normal;
    font-style: normal;
}}
@font-face {{
    font-family: "Computer Modern";
    src: url("{bold}");
    font-weight: bold;
    font-style: normal;
}}
@font-face {{
    font-family: "Computer Modern";
    src: url("{italic}");
    font-weight: normal;
    font-style: italic, oblique;
}}
@font-face {{
    font-family: "Computer Modern";
    src: url("{bi}");
    font-weight: bold;
    font-style: italic, oblique;
}}
'''
TITLE = "WCA Regulations"
AUTHOR = "WCA Regulations Committee"
HTML_TITLE = ('<div class="title">{title}</div>'
              '<div class="author">{author}</div>'
              '<div class="spacer"></div>')
NO_BREAK = '<div class="no_break_inside">'


class WCADocumentHtmlToPdf(WCADocumentHtml):
    ''' Emit html suitable to be printed to PDF using wkhtmltopdf '''
    name = "PDF"
    def __init__(self, versionhash, language, pdf):
        super(WCADocumentHtmlToPdf, self).__init__(versionhash, language, pdf)
        self.urls = {'regulations': '', 'guidelines': '',
                     'pdf': pdf}
        self.emit_rails_header = False
        self.emit_toc = False
        self.harticle = ('<div id="{anchor}"></div>'
                         '<h2 id="article-{anchor}-{new}"> '
                         '{name}{sep}{title}'
                         '</h2>\n')
        self.label = ('<li>[<span class="{name} label label-default">{name}</span>] '
                      '{text}</li>\n')
        self.guideline = ('<li id="{i}" class="rule">{i}) '
                          '<span class="{label} label {linked}">'
                          '[<a {attr}>{label}</a>]</span> {text}</li>\n')
        # Here we intentionally break the hierarchy (ul(li(ul(li))li()) turns to
        # ul(li()ul(li())li()) to be able to "easily" avoid  page breaking
        # inside a 'li' text (it does weird stuff if the whole element has to
        # avoid page-breaking
        self.regulation = '<li id="{i}" class="rule">{i}) {text}</li>\n'
        self.postreg = ''
        # Internal variable to handle grouping between a h* and its first element
        self.group_closed = True

    def generate_ul(self, a_list):
        # Here we don't want to generate 'ul' for the root rule, we want to group
        # the first 'li' with the header to force avoiding page-break
        return (len(a_list) > 0 and
                ((isinstance(a_list[0], Rule) and
                  not isinstance(a_list[0].parent, Article)) or
                 isinstance(a_list[0], LabelDecl)))

    def no_break(self):
        if self.group_closed:
            self.group_closed = False
            # This first div will be closed after the first paragraph
            self.codegen += NO_BREAK

    def visitArticle(self, article):
        self.no_break()
        return super(WCADocumentHtmlToPdf, self).visitArticle(article)

    def visitSection(self, section):
        self.no_break()
        return super(WCADocumentHtmlToPdf, self).visitSection(section)

    def visitSubsection(self, subsection):
        self.no_break()
        return super(WCADocumentHtmlToPdf, self).visitSubsection(subsection)

    def visitTableOfContent(self, toc):
        self.no_break()
        return super(WCADocumentHtmlToPdf, self).visitTableOfContent(toc)

    def visitstr(self, u):
        retval = super(WCADocumentHtmlToPdf, self).visitstr(u)
        if len(u) > 0 and not self.group_closed:
            # We are in an introduction paragraph and we want to group it
            # with the article's h2
            self.group_closed = True
            self.codegen += "</div>"
        return retval

    def visitRule(self, rule):
        # This one is called after the Rule's 'li' is emitted, we override
        # the call iterating through the Rule's children
        if not self.group_closed:
            self.group_closed = True
            self.codegen += "</div>"
        return super(WCADocumentHtmlToPdf, self).visitRule(rule)

    def visitWCARegulations(self, document):
        self.codegen += '<html><head>'
        self.codegen += '<title>%s</title>' % TITLE
        self.codegen += '<style media="all">\n'
        # Output the font faces for our custom fonts, shipped with the package
        fonts = {'normal': 'cmunrm.otf', 'italic': 'cmunti.otf',
                 'bold': 'cmunbx.otf', 'bi': 'cmunbi.otf'}
        for name in fonts.keys():
            fontfile = as_file(files("wrc").joinpath(f"data/{filename}"))
            if not os.path.isabs(fontfile):
                fontfile = os.path.abspath(fontfile)
            fonts[name] = fontfile
        self.codegen += CSS_FONTS.format(normal=fonts['normal'], bold=fonts['bold'],
                                         italic=fonts['italic'], bi=fonts['bi'])
        self.codegen += files("wrc").joinpath("data/htmltopdf.css").read_text(encoding="utf-8")
        self.codegen += '</style></head><body class="%s"><div>\n' % self.language
        self.codegen += HTML_TITLE.format(title=TITLE, author=AUTHOR)
        retval = super(WCADocumentHtmlToPdf, self).visitWCARegulations(document)
        self.codegen += '<div class="page_break"></div>\n'
        return retval

    def visitWCAGuidelines(self, document):
        retval = super(WCADocumentHtmlToPdf, self).visitWCAGuidelines(document)
        self.codegen += '</div></body></html>\n'
        return retval



''' Generic interface for a Backend '''
from wrc.sema.ast import ASTVisitor

class CGDocument(ASTVisitor):
    '''
    Provide a generic class to derive from when implementing a backend.
    Default codegen is a unicode string
    '''
    name = "undefined"
    def __init__(self, cg_type):
        super(CGDocument, self).__init__()
        self.cg_type = cg_type
        self.codegen = self.cg_type()

    def emit(self, ast_reg, ast_guide):
        ''' Default emit method: visit both ASTs and return the codegen '''
        if (ast_reg):
            self.visit(ast_reg)
        codegen_reg = self.codegen
        self.codegen = self.cg_type()
        if (ast_guide):
            self.visit(ast_guide)
        return (codegen_reg, self.codegen)

    def shouldEmitSubsection(self, language, title):
        return not (language == "english" and title == "Translation")



''' Backend for HTML. '''
import re
from wrc.sema.ast import Ruleset, Rule, LabelDecl
from wrc.codegen.cg import CGDocument

REPO_REG = "https://github.com/thewca/wca-regulations"
REPO_TRANS = "https://github.com/thewca/wca-regulations-translations"
BRANCH_REG = "official"
BRANCH_TRANS = "main"
ID_REG = "official"
ID_TRANS = "wca-regulations-translations"
TOC_ELEM = '<li>{name}{sep}<a href="#{anchor}">{title}</a></li>\n'
H2 = '<h2 id="{anchor}">{title}</h2>\n'
H3 = '<h3 id="{anchor}">{title}</h3>\n'
PROVIDE = "<% provide(:title, '{title}') %>\n"
TITLE = '<h1>{title}</h1>\n'
VERSION = '<div class="version">{version}<br/>{gitlink}</div>\n'
GITLINK = ('[<code><a href="{repo}/tree/{branch}/{gitdir}">{identifier}</a>'
           ':<a href="{repo}/commits/{version}">{version}</a></code>]')

# Some homemade basics non-robust md2html functions
# We could also call pandoc to do the conversion but it's awfully slow, and the
# md subset implemented right now is sufficient for the Regulations needs

def anchorizer(text):
    ''' Turn a text into an acceptable anchor name '''
    return text.lower().replace(" ", "-")


def special_links_replace(text, urls):
    '''
    Replace simplified Regulations and Guidelines links into actual links.
    'urls' dictionary is expected to provide actual links to the targeted
    Regulations and Guidelines, as well as to the PDF file.
    '''
    match_number = r'([A-Za-z0-9]+)' + r'(\+*)'
    reference_list = [(r'regulations:article:' + match_number, urls['regulations']),
                      (r'regulations:regulation:' + match_number, urls['regulations']),
                      (r'guidelines:article:' + match_number, urls['guidelines']),
                      (r'guidelines:guideline:' + match_number, urls['guidelines']),
                     ]
    anchor_list = [(r'regulations:contents', urls['regulations'] + r'#contents'),
                   (r'guidelines:contents', urls['guidelines'] + r'#contents'),
                   (r'regulations:top', urls['regulations'] + r'#'),
                   (r'guidelines:top', urls['guidelines'] + r'#'),
                   (r'link:pdf', urls['pdf'] + '.pdf'),
                  ]
    retval = text
    for match, repl in reference_list:
        retval = re.sub(match, repl + r'#\1\2', retval)
    for match, repl in anchor_list:
        retval = re.sub(match, repl, retval)
    return retval

def list2html(text):
    '''
    Very simple replacement for lists, no nesting, not even two lists in the
    same 'text'... (yet sufficient for the current regulations)
    Assumes list is in a paragraph.
    '''
    match = r'- (.+)\n'
    replace = r'<li>\1</li>\n'
    text = re.sub(match, replace, text)
    # Set start of list
    text = text.replace('<li>', '</p><ul><li>', 1)
    # Set end of list
    tmp = text.rsplit('</li>', 1)
    return '</li></ul><p>'.join(tmp)

def link2html(text):
    ''' Turns md links to html '''
    match = r'\[([^\]]+)\]\(([^)]+)\)'
    replace = r'<a href="\2">\1</a>'
    return re.sub(match, replace, text)

def simple_md2html(text, urls):
    ''' Convert a text from md to html '''
    retval = special_links_replace(text, urls)
    # Create a par break for double newlines
    retval = re.sub(r'\n\n', r'</p><p>', retval)
    # Create a visual br for every new line
    retval = re.sub(r'\n', r'<br />\n', retval)
    # Do we really need this ? Help reduce the diff to only '\n' diff.
    retval = re.sub(r'"', r'&quot;', retval)
    retval = list2html(retval)
    return link2html(retval)


class WCADocumentHtml(CGDocument):
    ''' Emit html formatted to fit in the WCA website.  '''
    name = "HTML"

    def __init__(self, versionhash, language, pdf, merged=False):
        super(WCADocumentHtml, self).__init__(str)
        self.regset = set()
        if merged:
            self.urls = {'regulations': './', 'guidelines': './', 'pdf': pdf}
        else:
            self.urls = {'regulations': './', 'guidelines': './guidelines.html', 'pdf': pdf}
        self.language = language

        is_translation = (language != "english")
        repo = REPO_TRANS if is_translation else REPO_REG
        branch = BRANCH_TRANS if is_translation else BRANCH_REG
        gid = ID_TRANS if is_translation else ID_REG
        gdir = language if is_translation else ""

        self.gitlink = GITLINK.format(repo=repo, branch=branch, identifier=gid,
                                      version=versionhash, gitdir=gdir)
        # Overridable attributes
        self.emit_rails_header = True
        self.emit_toc = True
        self.harticle = ('<div id="{anchor}"><div id="{old}">'
                         '<h2 id="article-{anchor}-{new}"> '
                         '<a href="#article-{anchor}-{new}">{name}</a>{sep}{title}'
                         '</h2></div></div>\n')
        self.label = ('<li><span class="{name} label label-default">{name}</span> '
                      '{text}</li>\n')
        self.guideline = ('<li id="{i}"><a href="#{i}">{i}</a>) '
                          '<span class="{label} label {linked}">'
                          '<a {attr}>{label}</a></span> {text}</li>\n')
        self.regulation = '<li id="{i}"><a href="#{i}">{i}</a>) {text}'
        self.postreg = '</li>\n'

    def md2html(self, text):
        ''' Use the simple markdown to html converter with our URLs '''
        return simple_md2html(text, self.urls)

    def generate_ul(self, a_list):
        ''' Determines if we should generate th 'ul' around the list 'a_list' '''
        return len(a_list) > 0 and (isinstance(a_list[0], Rule) or
                                    isinstance(a_list[0], LabelDecl))

    def visitWCADocument(self, document):
        self.codegen += '<div class="container">'
        self.codegen += TITLE.format(title=document.title)
        self.codegen += VERSION.format(version=document.version,
                                       gitlink=self.gitlink)
        retval = [self.visit(s) for s in document.sections]
        self.codegen += "</div>\n"
        # FIXME do we really need ascii html entities instead of plain utf8 ?
        self.codegen = self.codegen.encode('ascii', 'xmlcharrefreplace').decode('utf-8')
        # Let's provide the title in utf8, Rails should be able to handle it
        if self.emit_rails_header:
            self.codegen = (PROVIDE.format(title=document.title)
                            + self.codegen)
        return retval.count(False) == 0

    def visitlist(self, o):
        genul = self.generate_ul(o)
        if genul:
            self.codegen += '\n<ul>\n'
        retval = super(WCADocumentHtml, self).visitlist(o)
        if genul:
            self.codegen += "</ul>\n"
        return retval

    def visitstr(self, u):
        if len(u) > 0:
            self.codegen += "<p>" + self.md2html(u) + "</p>\n"
        return True

    def visitTableOfContent(self, toc):
        self.codegen += H2.format(anchor="contents",
                                  title=self.md2html(toc.title))
        retval = super(WCADocumentHtml, self).visit(toc.intro)
        if self.emit_toc:
            self.codegen += '<p><ul id="table_of_contents">\n'
            for article in toc.articles:
                anchorname = "article-" + article.number + "-" + article.newtag
                self.codegen += TOC_ELEM.format(name=article.name, anchor=anchorname,
                                                title=article.title, sep=article.sep)
            self.codegen += '</ul>\n</p>\n'
        return retval

    def visitSection(self, section):
        self.codegen += H2.format(anchor=anchorizer(section.title),
                                  title=section.title)
        return super(WCADocumentHtml, self).visitSection(section)

    def visitArticle(self, article):
        self.codegen += self.harticle.format(anchor=article.number,
                                             old=article.oldtag,
                                             new=article.newtag,
                                             name=article.name,
                                             title=article.title,
                                             sep=article.sep)
        retval = super(WCADocumentHtml, self).visit(article.intro)
        retval = retval and super(WCADocumentHtml, self).visit(article.content)
        return retval

    def visitSubsection(self, subsection):
        if self.shouldEmitSubsection(self.language, subsection.title):
            self.codegen += H3.format(anchor=anchorizer(subsection.title),
                                      title=subsection.title)
            return super(WCADocumentHtml, self).visitSubsection(subsection)
        return True

    def visitRegulation(self, reg):
        self.codegen += self.regulation.format(i=reg.number,
                                               text=self.md2html(reg.text))
        retval = super(WCADocumentHtml, self).visitRegulation(reg)
        self.codegen += self.postreg
        return retval


    def visitLabelDecl(self, decl):
        self.codegen += self.label.format(name=decl.name, text=decl.text)
        return True

    def visitGuideline(self, guide):
        reg = guide.regname
        linked = reg in self.regset
        label_class = "linked" if linked else ""
        link_attr = 'href="%s#%s"' % (self.urls['regulations'], reg)
        anchor_attr = 'id="#%s"' % guide.number
        attr = link_attr if linked else anchor_attr

        self.codegen += self.guideline.format(i=guide.number,
                                              text=self.md2html(guide.text),
                                              label=guide.labelname,
                                              linked=label_class,
                                              attr=attr)
        return super(WCADocumentHtml, self).visitGuideline(guide)

    def emit(self, ast_reg, ast_guide):
        self.regset = Ruleset().get(ast_reg)
        return super(WCADocumentHtml, self).emit(ast_reg, ast_guide)



from wrc.sema.ast import Article, Section, TableOfContent, Regulation, Guideline, split_rule_number
from copy import deepcopy

NOTES_INDEX = 0
TOC_INDEX = 1
FIRST_ARTICLE_INDEX = 2


class BadFormatError(Exception):
    def __init__(self, message: str):
        self.message = message + '. Please report this issue to the developer.'
        super().__init__(message)


def article_number_is_lower(num1: str, num2: str) -> bool:
    """
    Returns True if num1 < num2, otherwise returns False.
    This is necessary because:
    * We need to compare article numbers like '1' and 'A'.
    * '2' < '10' is False, which is not what we want.

    We want the following examples to evaluate to True:
    * '5' < '12'.
    * '12' < 'A'.
    * 'A' < 'B'.
    """
    is_digit_1 = num1.isdigit()
    is_digit_2 = num2.isdigit()
    if is_digit_1 and is_digit_2:
        # Both strings contain numbers.
        return int(num1) < int(num2)
    elif is_digit_1 and not is_digit_2:
        # Only num2 is a letter (always number < letter).
        return True
    elif not is_digit_1 and not is_digit_2:
        # Both are letters.
        return num1 < num2

    # The remaining case (num1 letter and num2 number) is always False.
    return False


def reg_number_is_lower(num1: list, num2: list) -> bool:
    """
    Returns True if num1 < num2, otherwise returns False.
    Note: this function is always used to compare regs that belong to the same article, so the order of numbers
    and letters is always the same for both num1 and num2.

    We want the following examples to evaluate to True:
    * [1, 'a'] < [1, 'a', 1] < [1, 'a', 2, 'b'] < [1, 'c']
    * ['B', 2] < ['B', 2, 'd'] < ['B', 2, 'd', 1] < ['B', 3]
    """
    for n1, n2 in zip(num1, num2):
        if n1 < n2:
            return True
        elif n1 > n2:
            return False

    # zip() includes the items up to the length of the shorter list.
    # We get here if the lengths are different but all the items included by zip are equal.
    # For example, if we are comparing: [2, 'a'] < [2, 'a', 1]. The for loop compares up to [2, 'a'], and as it does not
    # return from there, we just check if num1 had a shorter list than num2.
    return len(num1) < len(num2)


def get_reg_number(reg) -> list:
    if isinstance(reg, Guideline):
        reg = reg.number.split('+')[0]
        num = split_rule_number(reg)
    elif isinstance(reg, Regulation):
        num = split_rule_number(reg.number)
    else:
        raise BadFormatError(f'get_reg_number expected Regulation or Guideline, got {type(reg).__name__}')
    if str(num[-1]) == '0':
        # Because split_rule_number() sometimes returns a 0 at the end, and we don't want that here.
        del num[-1]
    return num


def merge_ast(astreg, astguide, language_options):
    """
    Combine the two AST into one.
    """
    # We only want the Articles, so we will start the for loop below at FIRST_ARTICLE_INDEX,
    # and check if the indices for Notes, TOC and First Article are correct.
    if not isinstance(astreg.sections[NOTES_INDEX], Section) or \
            not isinstance(astguide.sections[NOTES_INDEX], Section):
        raise BadFormatError('NOTES_INDEX is incorrect')
    if not isinstance(astreg.sections[TOC_INDEX], TableOfContent):
        raise BadFormatError('TOC_INDEX is incorrect')
    for document in [astreg, astguide]:
        if not hasattr(document.sections[FIRST_ARTICLE_INDEX], 'number') or \
                document.sections[FIRST_ARTICLE_INDEX].number != '1':
            raise BadFormatError('FIRST_ARTICLE_INDEX is incorrect')

    # We take the astreg as the base, and then we add the guidelines in their corresponding places.
    ast_combined = deepcopy(astreg)

    # Merge the notes:
    # We won't include the "WCA Regulations" notes from the Guidelines, as it is the same as in
    # "WCA Regulations and Guidelines" of the Regulations document.
    # languages_options is used to handle translations.
    ast_combined.sections[NOTES_INDEX].content.append(
        [cont for cont in astguide.sections[NOTES_INDEX].content if language_options['regulations'] not in cont.title])

    # Now we iterate over all the Articles found in astguide, from FIRST_ARTICLE_INDEX until the end of the
    # sections list.
    guide_sections = astguide.sections[FIRST_ARTICLE_INDEX:]
    for guide_section in guide_sections:

        stack = deepcopy(guide_section.content)
        reg_section = None
        # Find the associated section in the Regulations or add it if there isn't one.
        ast_combined_sections = ast_combined.sections[FIRST_ARTICLE_INDEX:]
        for index, section in enumerate(ast_combined_sections):
            if not isinstance(section, Article):
                # We could just continue here, but finding a non-Article section after FIRST_ARTICLE_INDEX
                # is worth raising an exception.
                raise BadFormatError('Section is not an Article')

            if section.number == guide_section.number:
                # We found the section!
                reg_section = section
                break
            elif article_number_is_lower(guide_section.number, section.number):
                # This happens when an Article in the Guidelines doesn't have an associated article in the Regulations.
                ast_combined.sections[TOC_INDEX].articles.insert(index, guide_section)
                # Now we copy the Article from the Guidelines into the ast_combined and continue (empty stack).
                ast_combined.sections.insert(index + FIRST_ARTICLE_INDEX, guide_section)
                stack = []
                break

        # Insert guidelines until the stack is empty.
        while len(stack) > 0:
            guideline = stack.pop(0)
            guideline_num = get_reg_number(guideline)
            inserted_index = recursive_insert(reg_section.content, guideline, guideline_num, stack)
            if inserted_index == -1:
                raise BadFormatError(f'Could not insert guideline {guideline.number}')

    return ast_combined


def recursive_insert(root, guideline, guideline_num, stack) -> int:
    """
    Insert the guideline using a recursive algorithm.
    """
    inserted_index = -1  # We use this as a flag and as an auxiliary variable to insert.
    for index, node in enumerate(root):
        if inserted_index != -1:
            # The guideline was inserted in the previous iteration.
            break
        if hasattr(node, 'siblings') and node.siblings:
            # The 'siblings' boolean attribute is used to insert after the guideline with the most pluses (+).
            continue

        node_num = get_reg_number(node)

        # if len == len and node_num >= guideline_num.
        if len(node_num) == len(guideline_num) and not reg_number_is_lower(node_num, guideline_num):
            if node_num == guideline_num:
                # The guideline goes below the current node.
                inserted_index = index + 1
            elif reg_number_is_lower(guideline_num, node_num):
                # We have len == len, but the iteration went over the number without finding a regulation with the same
                # number. The guideline doesn't have an associated regulation, so we insert it before the current node.
                inserted_index = index

            # Insert:
            node.siblings = True
            root.insert(inserted_index, guideline)

            # Now we look if we have more guidelines for the same regulation in the stack, and we add all of them
            # together.
            while len(stack) > 0 and get_reg_number(stack[0]) == guideline_num:
                plus_guideline = stack.pop(0)
                root[inserted_index].siblings = True
                inserted_index += 1
                root.insert(inserted_index, plus_guideline)

            # Move the children of the affected Regulation to the children of the last Guideline added
            # for such Regulation.
            root[inserted_index].children = node.children
            node.children = []
            break

        # The first part of the numeration matches, but the numeration of the guideline is longer, so we go down
        # with the current node as the root.
        elif len(node_num) < len(guideline_num) and node_num == guideline_num[:len(node_num)]:
            inserted_index = recursive_insert(node.children, guideline, guideline_num, stack)

        # else: next node.

    if inserted_index == -1:
        # If we couldn't insert the guideline anywhere, we insert it at the end of the root's content.
        root.append(guideline)
        inserted_index = len(root) - 1

    return inserted_index



'''
WCALexer : tokenize a string using lex.
'''
import re
import ply.lex as lex
from unidecode import unidecode


class WCALexer(object):

    # Token names in the Regulations and Guidelines
    tokens = (
        'TITLE',
        'VERSION',
        'STATESTAG',
        'TOC',
        'LABELDECL',
        'ARTICLEHEADER',
        'STATESHEADER',
        'HEADERSEC',
        'HEADERSUBSEC',
        'TEXT',
        'PARBREAK',
        'REGULATION',
        'GUIDELINE',
        'STATE',
        )


    t_ignore = '\t'

    def t_TITLE(self, token):
        r'\#\s+<wca-title>(?P<title>.+)\n'
        token.value = token.lexer.lexmatch.group("title")
        token.lexer.lineno += 1
        return token

    def t_VERSION(self, token):
        r'<version>(?P<version>.+)\n'
        token.lexer.lineno += 1
        token.value = token.lexer.lexmatch.group("version")
        return token

    def t_STATESTAG(self, token):
        r'<wca-states>\n'
        token.lexer.lineno += 1
        token.value = None
        return token

    def t_TOC(self, token):
        r'<table-of-contents>'
        token.value = token.value
        return token

    def t_LABELDECL(self, token):
        r'-\s<label>\s*\[(?P<label>.+?)\]\s*(?P<text>.+?)\n'
        label = token.lexer.lexmatch.group("label")
        text = token.lexer.lexmatch.group("text")
        token.value = (label, text)
        token.lexer.lineno += 1
        return token

    def t_ARTICLEHEADER(self, token):
        # In the "sep" group, there is a distinct difference between `:` (ASCII colon) and `：` (CJK full-width colon)
        r'\#\#\s+<article-(?P<number>[A-Z0-9]+)><(?P<newtag>[a-zA-Z0-9-]+)><(?P<oldtag>[a-zA-Z0-9-]+)>[ ]*(?P<name>[^\<]+?)(?P<sep>:\s|：)(?P<title>[^<\n]+)\n'
        number = token.lexer.lexmatch.group("number")
        newtag = token.lexer.lexmatch.group("newtag")
        oldtag = token.lexer.lexmatch.group("oldtag")
        name = token.lexer.lexmatch.group("name")
        sep = token.lexer.lexmatch.group("sep")
        title = token.lexer.lexmatch.group("title")
        token.value = (number, newtag, oldtag, name, title, sep)
        token.lexer.lineno += 1
        return token

    def t_STATESHEADER(self, token):
        r'\#\#\s+<states-list>(?P<title>[^<\n]*)\n'
        title = token.lexer.lexmatch.group("title")
        token.value = title
        token.lexer.lineno += 1
        return token

    def t_HEADERSEC(self, token):
        r'\#\#\s+(?P<title>.+?)\n'
        title = token.lexer.lexmatch.group("title")
        token.value = title
        token.lexer.lineno += 1
        return token

    # This is not very flexible, but make the yacc very straightforward
    def t_HEADERSUBSEC(self, token):
        r'\#\#\#\s+(?P<title>.+?)\n'
        title = token.lexer.lexmatch.group("title")
        token.value = title
        token.lexer.lineno += 1
        return token

    def t_REGULATION(self, token):
        r'(?P<indents>\s{4,})*-\s(?P<reg>[a-zA-Z0-9]+)\)\s*(?P<text>.+?[^ ])\n'
        indents = token.lexer.lexmatch.group("indents")
        indents = len(indents)//4 if indents else 0
        reg = token.lexer.lexmatch.group("reg")
        text = token.lexer.lexmatch.group("text")
        token.value = (indents, reg, text)
        token.lexer.lineno += 1
        return token

    def t_GUIDELINE(self, token):
        r'(?P<indents>\s{4,})*-\s(?P<reg>[a-zA-Z0-9]+[+]+)\)\s\[(?P<label>.+?)\]\s*(?P<text>.+?[^ ])\n'
        indents = token.lexer.lexmatch.group("indents")
        indents = len(indents)//4 if indents else 0
        reg = token.lexer.lexmatch.group("reg")
        text = token.lexer.lexmatch.group("text")
        label = token.lexer.lexmatch.group("label")
        token.value = (indents, reg, text, label)
        token.lexer.lineno += 1
        return token

    def t_STATE(self, token):
        r'-\s\((?P<state>[A-Z]{2}):(?P<continent>[_A-Za-z ]+)(:(?P<friendly_id>[A-Za-z_]+))?\)\s(?P<name>[A-Z].+?[^ ])\n'
        state = token.lexer.lexmatch.group("state")
        continent = token.lexer.lexmatch.group("continent")
        name = token.lexer.lexmatch.group("name")
        friendly_id = token.lexer.lexmatch.group("friendly_id")
        if friendly_id:
            friendly_id = friendly_id
        else:
            friendly_id = unidecode(name).replace("'", "_")
        token.value = (state, continent, name, friendly_id)
        token.lexer.lineno += 1
        return token

    def t_TEXT(self, token):
        r'(?P<text>[^<#\n ].+?[^ ])(?=\n)'
        text = token.lexer.lexmatch.group("text")
        token.value = text
        return token

    def t_PARBREAK(self, token):
        r'\n{2,}'
        token.lexer.lineno += len(token.value)
        return token

    def t_trailingwhitespace(self, token):
        r'.+? \n'
        print("Error: trailing whitespace at line %s in text '%s'" % (token.lexer.lineno + 1, token.value[:-1]))
        token.lexer.lexerror = True
        token.lexer.skip(1)

    def t_newline(self, token):
        r'\n+'
        token.lexer.lineno += len(token.value)

    def t_error(self, token):
        print("Illegal character '%s' at line %s" % (token.value[0], token.lexer.lineno))
        token.lexer.lexerror = True
        token.lexer.skip(1)

    def lex(self):
        return lex.lex(module=self, reflags=re.UNICODE, debug=0)


